## Rosbrowz

This library provides an rviz like interface for the browser with minimal effort.

Typescript bindings are included in the library.

#### Installation

```bash
npm install rosbrowz --save
```

#### Usage

##### Initialization

```typescript
import Rosbrowz from "rosbrowz";

const r2d = new Rosbrowz.R2d({
  container: document.getElementById("rosbrowz-holder")!,
  zoom: true,
  pan: true,
  scale: 15,
  ros: {
    url: "ws://localhost:9090"
  }
});
```

The parameter `ros` can either be:

```typescript
ros: {
  url: <ros-bridge-url>
}
```

or

```typescript
ros: new ROSLIB.Ros({
  url: <ros-bridge-url>
})
```

The second one is useful in case the user wants to use a `RosLIB.Ros` instance across multiple places.

##### Creating a background map

```typescript
const map = await Rosbrowz.R2d.createMap({
  path: require("../map.png") // or a url,
  info: { 
    origin: [-43.6567, -121.957],
    resolution: 0.10000000149
  }
});

r2d.addMap(map);
```

##### Creating objects

```typescript
const circle = Rosbrowz.R2d.createCircle({
  radius: 1,
  color: "yellow",
  border: 0.2,
  borderColor: "black",
  opacity: 0.9
});

const rectangle = Rosbrowz.R2d.createRectangle({
  width: 2,
  height: 1,
  color: "yellow",
  border: 0.2,
  borderColor: "black",
  opacity: 1
});

const arrow = Rosbrowz.R2d.createArrow({
  pointerWidth: 1.5,
  pointerLength: 2,
  color: "yellow",
  border: 0.2,
  borderColor: "black",
  opacity: 1
});

const image = (await Rosbrowz.R2d.createImageAsync({
  imageURL: require("../forklift.png"),
  height: 3 / 3.5,
  width: 3
})).image;
```

Note that these objects are floating and just creating them does not mean that they will be rendered in the map. These floating objects are supposed to be used in other contexts like `addPose` and `addStaticObject` as below.

##### Adding a pose

```typescript
r2d.addPose({
  object: image,
  topic: "/slam/pose"
});
```

##### Adding a static object

```typescript
r2d.addStaticObject({
  object: circle,
  x: -10,
  y: -10,
  rotationInDegClockwise: 90
});
```

##### Adding a path

```typescript
r2d.addPath({
  points: [
    { x: 0, y: 0 },
    { x: -10, y: -10 },
    { x: -15, y: -5 },
    { x: -20, y: -15 },
    { x: -10, y: -20 }
  ],
  width: 0.5,
  color: "red",
  tension: 1
});
```

#### Removing objects

Each of `addPose`, `addStaticObject`, `addPath` return an object with a `destroy` function that can be called to remove them from the canvas.

Example:

```typescript
const poseController = r2d.addPose({
  object: image,
  topic: "/slam/pose"
});

poseController.destroy();
```
