import * as Konva from "konva";
import * as ROSLib from "roslib";
import {
  animateToPoint,
  getContainerDimensions,
  isRosBridgeURL,
  quaternionToEulerAngle,
  ZoomControl,
  isObjectRectangle,
  isObjectImage
} from "./utils";
import circle from "./helpers/circle";
import { isNil, map } from "lodash";
import rectangle from "./helpers/rectangle";
import arrow from "./helpers/arrow";
import { imageAsync, imageSync } from "./helpers/image";
import drawableMap from "./helpers/map";

interface IR2d {
  container: HTMLElement;
  zoom: boolean;
  pan: boolean;
  scale: number;
  ros: IRosBridgeURL | IRos;
}

interface IMap {
  image: HTMLImageElement;
  info: {
    origin: number[];
    resolution: number;
  };
}

interface IPose {
  object: Konva.Circle | Konva.Rect | Konva.Arrow | Konva.Image;
  topic: string;
  queueLength?: number;
  throttleRate?: number;
}

interface IHeader {
  frame_id: "map";
  seq: number;
  stamp: {
    secs: number;
    nsecs: number;
  };
}

interface Ixyz {
  x: number;
  y: number;
  z: number;
}

interface IPoseMsg {
  header: IHeader;
  pose: {
    position: Ixyz;
    orientation: Ixyz & {
      w: number;
    };
  };
}

export interface IRosBridgeURL {
  url: string;
}

export type IRos = ROSLib.Ros;

interface IStaticObject {
  object: Konva.Circle | Konva.Rect | Konva.Arrow | Konva.Image;
  x: number;
  y: number;
  rotationInDegClockwise?: number;
}

interface IPath {
  points: Array<{ x: number; y: number }>;
  color: string;
  width: number;
  tension?: number;
}

export default class R2d {
  private readonly container: HTMLElement;
  private readonly stage: Konva.Stage;
  private readonly zoom: boolean;
  private readonly pan: boolean;
  private readonly scale: number;
  private readonly ros: ROSLib.Ros;
  private readonly mapLayer: Konva.Layer;
  private readonly objectsLayer: Konva.Layer;

  static createMap = drawableMap;
  static createCircle = circle;
  static createRectangle = rectangle;
  static createArrow = arrow;
  static createImageSync = imageSync;
  static createImageAsync = imageAsync;

  constructor(args: IR2d) {
    this.container = args.container;
    this.zoom = args.zoom;
    this.pan = args.pan;
    this.scale = args.scale;
    this.stage = new Konva.Stage({
      container: this.container,
      ...getContainerDimensions(this.container),
      draggable: this.pan
    });
    this.ros = !isRosBridgeURL(args.ros)
      ? args.ros
      : new ROSLib.Ros({ url: args.ros.url });
    this.mapLayer = new Konva.Layer();
    this.objectsLayer = new Konva.Layer();

    this.stage.add(this.mapLayer);
    this.stage.add(this.objectsLayer);

    if (this.zoom) {
      const zoomControl = new ZoomControl(this.stage);
      zoomControl.handleWheelZoom(this.container, 1.1);
      zoomControl.handleTouchZoom();
    }
  }

  public addMap = (map: IMap) => {
    const { image, info } = map;
    const { origin, resolution } = info;

    const konvaImage = new Konva.Image({
      image,
      scaleX: resolution * this.scale,
      scaleY: resolution * this.scale,
      x: this.stage.width() / 2 + origin[0] * this.scale,
      y:
        this.stage.height() / 2 -
        image.naturalHeight * resolution * this.scale -
        origin[1] * this.scale
    });
    this.mapLayer.add(konvaImage);
    this.mapLayer.batchDraw();

    return {
      destroy: () => {
        konvaImage.destroy();
        this.mapLayer.batchDraw();
      }
    };
  };

  public addPose = (pose: IPose) => {
    const listenerArg: {
      ros: ROSLib.Ros;
      name: string;
      messageType: string;
      queue_length?: number;
      throttle_rate?: number;
    } = {
      ros: this.ros,
      name: pose.topic,
      messageType: "geometry_msgs/PoseStamped"
    };
    !isNil(pose.queueLength)
      ? (listenerArg.queue_length = pose.queueLength)
      : null;
    !isNil(pose.throttleRate)
      ? (listenerArg.throttle_rate = pose.throttleRate)
      : null;

    const listener = new ROSLib.Topic(listenerArg);
    pose.object.scaleX(this.scale);
    pose.object.scaleY(this.scale);
    this.objectsLayer.add(pose.object);
    pose.object.hide();

    listener.subscribe((message: IPoseMsg) => {
      pose.object.show();
      const { x: moveX, y: moveY } = message.pose.position;
      const { gamma } = quaternionToEulerAngle(message.pose.orientation);
      animateToPoint(
        pose.object,
        this.stage.width() / 2 + moveX * this.scale,
        this.stage.height() / 2 - moveY * this.scale
      );
      const currentAngle = pose.object.rotation() % 360;
      const desiredAngle =
        ((gamma > 0 ? gamma : 2 * Math.PI + gamma) * 180) / Math.PI;
      pose.object.rotate(-1 * (desiredAngle - Math.abs(currentAngle)));
      this.objectsLayer.batchDraw();
    });

    return {
      destroy: () => {
        pose.object.destroy();
        listener.unsubscribe();
        this.objectsLayer.batchDraw();
      }
    };
  };

  public addStaticObject = (args: IStaticObject) => {
    const { object, x, y, rotationInDegClockwise } = args;

    if (isObjectRectangle(object) || isObjectImage(object)) {
      object.offsetY(object.height() / 2);
      object.offsetX(object.width() / 2);
    }
    object.scaleX(this.scale);
    object.scaleY(this.scale);
    object.x(this.stage.width() / 2 - x * this.scale);
    object.y(this.stage.height() / 2 + y * this.scale);
    !isNil(rotationInDegClockwise)
      ? object.rotate(rotationInDegClockwise)
      : null;

    this.objectsLayer.add(object);
    this.objectsLayer.batchDraw();

    return {
      destroy: () => {
        object.destroy();
        this.objectsLayer.batchDraw();
      }
    };
  };

  public addPath = (args: IPath) => {
    const { points: pointsUnformatted, color, width } = args;
    const points: number[] = [];
    map(pointsUnformatted, p => {
      points.push(this.stage.width() / 2 - p.x * this.scale);
      points.push(this.stage.height() / 2 + p.y * this.scale);
    });

    const line = new Konva.Line({
      points,
      stroke: color,
      strokeWidth: width * this.scale,
      x: this.stage.width() / 2,
      y: this.stage.height() / 2,
      offset: {
        x: this.stage.width() / 2,
        y: this.stage.height() / 2
      },
      tension: !isNil(args.tension) ? args.tension : 0
    });

    this.objectsLayer.add(line);
    this.objectsLayer.batchDraw();

    return {
      destroy: () => {
        line.destroy();
        this.objectsLayer.batchDraw();
      }
    };
  };
}
