import * as Konva from "konva";
import { IRos, IRosBridgeURL } from "./index";
import { isNil, isEqual } from "lodash";

interface IPoint {
  x: number;
  y: number;
}

export const getContainerDimensions = (container: HTMLElement) => {
  const { width, height } = container.getBoundingClientRect();
  return { width, height };
};

export const zoomStage: (
  container: HTMLElement,
  stage: Konva.Stage,
  scaleBy: number
) => void = (container, stage, scaleBy) => {
  const oldScale = stage.scaleX();
  const { width, height } = getContainerDimensions(container);
  const mousePointTo = {
    x:
      (stage.getPointerPosition() || { x: width / 2 }).x / oldScale -
      stage.x() / oldScale,
    y:
      (stage.getPointerPosition() || { y: height / 2 }).y / oldScale -
      stage.y() / oldScale
  };

  const newScale = oldScale * scaleBy;
  stage.scale({ x: newScale, y: newScale });

  const newPos = {
    x:
      -(
        mousePointTo.x -
        (stage.getPointerPosition() || { x: width / 2 }).x / newScale
      ) * newScale,
    y:
      -(
        mousePointTo.y -
        (stage.getPointerPosition() || { y: height / 2 }).y / newScale
      ) * newScale
  };
  stage.position(newPos);
  stage.batchDraw();
};

const getDistance: (p1: IPoint, p2: IPoint) => number = (p1, p2) => {
  return Math.sqrt(Math.pow(p2.x - p1.x, 2) + Math.pow(p2.y - p1.y, 2));
};

export class ZoomControl {
  private readonly stage: Konva.Stage;
  private lastDist = 0;

  public constructor(stage: Konva.Stage) {
    this.stage = stage;
  }

  public handleWheelZoom(element: HTMLElement, scaleBy: number) {
    element.addEventListener("wheel", e => {
      e.preventDefault();
      zoomStage(element, this.stage, e.deltaY < 0 ? scaleBy : 1 / scaleBy);
    });
  }

  public handleTouchZoom() {
    this.stage.getContent().addEventListener(
      "touchmove",
      e => {
        const touch1 = e.touches[0];
        const touch2 = e.touches[1];

        if (touch1 && touch2 && this.stage) {
          const dist = getDistance(
            {
              x: touch1.clientX,
              y: touch1.clientY
            },
            {
              x: touch2.clientX,
              y: touch2.clientY
            }
          );

          if (!this.lastDist) {
            this.lastDist = dist;
          }

          const scale = (this.stage.scaleX() * dist) / this.lastDist;

          this.stage.scaleX(scale);
          this.stage.scaleY(scale);
          this.stage.draw();
          this.lastDist = dist;
        }
      },
      false
    );

    this.stage.getContent().addEventListener(
      "touchend",
      () => {
        this.lastDist = 0;
      },
      false
    );
  }
}

export const isRosBridgeURL = (
  object: IRos | IRosBridgeURL
): object is IRosBridgeURL => {
  return !isNil((<IRosBridgeURL>object).url);
};

export const isObjectRectangle = (
  object: Konva.Circle | Konva.Rect | Konva.Arrow | Konva.Image
): object is Konva.Rect => {
  return !isNil((<Konva.Rect>object).cornerRadius);
};

export const isObjectImage = (
  object: Konva.Circle | Konva.Rect | Konva.Arrow | Konva.Image
): object is Konva.Image => {
  return !isNil((<Konva.Image>object).image);
};

export const quaternionToEulerAngle = ({
  w,
  x,
  y,
  z
}: {
  w: number;
  x: number;
  y: number;
  z: number;
}) => {
  const t0 = 2.0 * (w * x + y * z);
  const t1 = 1.0 - 2.0 * (x * x + y * y);
  const alpha = Math.atan2(t0, t1);

  let t2 = 2.0 * (w * y - z * x);
  if (t2 > 1.0) {
    t2 = 1.0;
  } else if (t2 < -1.0) {
    t2 = -1.0;
  }
  const beta = Math.asin(t2);

  const t3 = 2.0 * (w * z + x * y);
  const t4 = 1.0 - 2.0 * (y * y + z * z);
  const gamma = Math.atan2(t3, t4);

  return { alpha, beta, gamma };
};

export const animateToPoint = (
  object: Konva.Circle | Konva.Rect | Konva.Arrow | Konva.Image,
  x: number,
  y: number
) => {
  object.to({
    duration: 0.1,
    easing: Konva.Easings.EaseIn,
    x,
    y
  });
};

export const listenForContainerResize = (
  previousDimensions: { width: number; height: number },
  container: HTMLElement,
  callback: (newDimensions: { width: number; height: number }) => void
) => {
  setTimeout(() => {
    const { width, height } = container.getBoundingClientRect();
    const dimensions = { width, height };
    if (!isEqual(previousDimensions, dimensions)) {
      callback(dimensions);
    }
    listenForContainerResize(dimensions, container, callback);
  }, 100);
};
